""" Base controller class for superwise package """
import json
import traceback
from abc import ABC
from abc import ABCMeta
from abc import abstractmethod
from abc import abstractproperty

from superwise import logger
from superwise.controller.client import Client
from superwise.controller.exceptions import *
from superwise.models.data_entity import DataEntity
from superwise.models.role import Role
from superwise.models.segment import Segment
from superwise.models.task import Task
from superwise.models.validation_error import ValidationError
from superwise.models.version import Version


class BaseController(ABC):
    """Base class for controllers"""

    __metaclass__ = ABCMeta

    def __init__(self, client, sw):
        """

        :param client: instance of client object (wrapper of requests)
        """
        self.client = client
        self.path = None
        self.model_name = abstractproperty()
        self.logger = logger
        self.model = None
        self.sw = sw

    def _dict_to_model(self, params, model_name=None):
        """

        :param params: dictionary of parameters
        :return: model
        """
        model_name = model_name or self.model_name
        try:
            if "self" in params:
                del params["self"]
            if isinstance(params, list):
                model = list()
                for param in params:
                    cmodel = globals()[model_name](**param)
                    model.append(cmodel)
            else:
                if "__class__" in params:
                    del params["__class__"]
                model = globals()[model_name](**params)
        except Exception as err:
            traceback.print_exc()
            raise Exception("exception in create {}".format(err))
        return model

    def post(self):
        """
        Prepare data and call run API POST call

        :return: requests response object
        """
        params = self.model.get_properties()
        self.logger.info("POST %s ", self.path)
        response = self.client.post(self.build_url(self.path), params)
        return response

    def patch(self, path=None, params=None):
        """
        Prepare data and call run API PATCH call

        :return: requests response object
        """
        path = path or self.path
        params = params or self.model.get_properties()
        self.logger.info("PATCH %s ", self.path)
        res = self.client.patch(self.build_url(path), params)
        return res

    def parse_response(self, response, model_name=None, is_return_model=True):
        """
        Format the response, change it from dict to be model based
        :param r: respnse object (created by requests package)
        :return:
        """
        model_name = model_name or self.model_name
        try:
            body = json.loads(response.content)
        except Exception as ex:
            raise Exception("error loading json, status code {} text {}".format(response.status_code, response.content))
        if response.status_code in [200, 201, 204]:
            res = self._dict_to_model(body, model_name=model_name) if is_return_model else body
        elif response.status_code in [500]:
            response.raise_for_status()
        else:
            res = ValidationError(http_status_code=response.status_code, http_error_reason=response.reason, body=body)
        return res

    def _create_update(self, model, is_return_model=True, create=True, model_name=None, **kwargs):
        """
        create object from an instance of model
        :param model:
        :return:
        """
        action = "create" if create else "update"
        try:
            if model.__class__.__name__ == self.model_name:
                self.model = model
                res = self.post() if create else self.patch()
            else:
                raise Exception(
                    "Model {} passed instead of {} to {}".format(
                        model.__class__.__name__, self.model_name, self.__class__.__name__
                    )
                )

            return self.parse_response(res, is_return_model=is_return_model)
        except SuperwiseValidationException:
            raise
        except Exception as e:
            traceback.print_exc()
            raise Exception("exception in {} {}".format(action, e))

    def create(self, model, return_model=True, **kwargs):
        """
        create object from an instance of model
        :param model:
        :return:
        """
        return self._create_update(model, create=True, return_model=return_model, **kwargs)

    def get(self, fields):
        """
        wrapper for reqeusts.get()

        :param fields: optinal fields for filtering
        :return:  response body string from backend (json)
        """
        self.logger.info("GET %s ", self.path)
        res = self.client.get(self.build_url(self.path), json=fields)
        return res.content

    def get_by_id(self, idx):
        """
        get a specifc resource by id
        :param id:
        :return:
        """
        url = self.build_url("{}/{}".format(self.path, idx))
        self.logger.info("GET  %s ", url)
        res = self.client.get(url)
        return self.parse_response(res)

    def build_url(self, path):
        """
        Build a url for a given path
        :param path: relative path, normally declared in each model (subclasses of this class)

        :return: URL of a resource
        """
        return "https://{}/{}/{}".format(self.client.api_host, self.client.tenant_id, path)

    def update(self, model, **kwargs):
        """

        :param model: model (instance of any subclass of BaseModel) to update

        """
        return self._create_update(model, create=False, **kwargs)



import os


def get_bool(key, default):
    v = os.environ.get(key, default)
    if v in ["False", "0", False]:
        return False
    else:
        return True


class Config:
    ENVIRONMENT = os.environ.get("ENVIRONMENT", "production")
    FRONTEGG_URL = os.environ.get("FRONTEGG_SUPERWISE_URL", "https://auth.superwise.ai")
    SUPERWISE_HOST = os.environ.get("SUPERWISE_HOST", "portal.superwise.ai")
    POOLING_INTERVAL_SEC = 15
    LIST_DROP_DATA_COLS = ["task_id", "version_id"]



import json
import os.path
import pathlib
from pprint import pprint

from superwise import Superwise
from superwise.models.data_entity import DataEntity
from tests.conf.config import config


def get_sw():
    """
    Get instance of superwise object for testing

    :return: superwise object
    """
    sw = Superwise(client_id=config["CLIENT_ID"], secret=config["SECRET"])
    return sw


def print_results(title, obj):
    print("####### {} #######".format(title))
    pprint(obj)
    print("\n")


def get_entities_fixture():
    with open("tests/resources/basic_schema.json", "r") as fh:
        schema = json.loads(fh.read())

    entities = [
        DataEntity(
            name=m["name"], type=m["type"], is_dimension=m["is_dimension"], role=m["role"], feature_importance=None
        )
        for m in schema
    ]
    return entities


COMMON_ASSETS_DIRECTORY = os.path.join(pathlib.Path(__file__).parent.absolute(), "common_assets")


def match_with_json_file(path, obj):
    with open(path, "r") as f:
        expected_result = json.load(f)
    assert expected_result == obj



"""
General functionallity tests
"""
import json
import os
from pprint import pprint

import pytest

from superwise import Superwise
from superwise.controller.exceptions import *
from superwise.models.task import Task
from tests import get_sw
from tests import print_results
from tests.conf.config import config


@pytest.mark.vcr()
def test_load_superwise_login_user_password():
    ok = True
    try:
        email = config.get("EMAIL", "email")
        password = config.get("PASSWORD", "email")
        print("login with {} and pass {}".format(email, password))
        sw = Superwise(email=email, password=password)
    except:
        ok = False
    assert ok is True


@pytest.mark.vcr()
def test_load_superwise_object_login():
    ok = True
    try:
        sw = Superwise(client_id=config["CLIENT_ID"], secret=config["SECRET"])
    except Exception as e:
        print(e)
        print("remnoivew vasdvasdvas")
        ok = False
    assert ok is True


@pytest.mark.vcr()
def test_invalid_token():
    ok = False
    try:
        sw = Superwise(client_id="test", secret="test_fail")
    except SuperwiseAuthException:
        ok = True
    assert ok is True


@pytest.mark.vcr()
def test_send_wrong_sdkmodel():
    sw = get_sw()
    model = Task()
    ok = False
    try:
        sw.version.create(model)
    except Exception as e:
        ok = True
    assert ok is True



""" This module contain BaseModel class  """
import datetime
from enum import Enum

import pandas as pd


class BaseModel:
    """ Base model """

    def __init__(self, **kwargs):
        self.id = kwargs("id", None)

    def from_datetime(self, date):
        if pd.isna(date):
            return None
        elif isinstance(date, datetime.datetime):
            return date.isoformat()
        else:
            return date

    def get_properties(self):

        """
        get properrties of model as dictionary

        :return: properties dict
        """
        return dict(
            (name, getattr(self, name))
            for name in dir(self)
            if not name.startswith("__") and not callable(getattr(self, name))
        )

    def get_enum_value(self, v):
        if isinstance(v, Enum):
            return v.value
        else:
            return v



import pytest

from superwise.controller.exceptions import *
from superwise.models.segment import Segment
from superwise.models.segment import SegmentCondition
from superwise.models.segment import SegmentConditionDefinition
from superwise.models.task import Task
from tests import get_sw
from tests import print_results

segment_id = None


@pytest.mark.vcr()
def test_create_segment_inline():
    sw = get_sw()
    inline_model_test = sw.segment.create(
        Segment(
            task_or_id=Task(id=1),
            name="My Segment #1",
            definition_or_json=[
                SegmentConditionDefinition(
                    entity_or_name="merchant_id",
                    condition=SegmentCondition.IN,
                    value=["Israel", "United States of America"],
                ),
                SegmentConditionDefinition(
                    entity_or_name="periodic_11634", condition=SegmentCondition.GREATER_THAN, value=0.5
                ),
            ],
        )
    )
    print_results("created segment object 1", inline_model_test.get_properties())
    assert inline_model_test.task_id == 1
    assert inline_model_test.name == "My Segment #1"
    assert inline_model_test.id is not None


@pytest.mark.vcr()
def test_create_segment():
    sw = get_sw()
    segment = Segment()
    global segment_id
    segment.task_id = 1
    segment.name = "My Segment #2"
    segment.definition_json = [
        {"entity_name": "merchant_id", "condition": "in", "value": ["Israel", "United States of America"]},
        {"entity_name": "periodic_11634", "condition": ">", "value": 0.5},
    ]
    new_segment_model = sw.segment.create(segment)
    print_results("created segment object 2", new_segment_model.get_properties())
    assert new_segment_model.name == "My Segment #2"
    assert new_segment_model.task_id == 1
    assert new_segment_model.id is not None
    segment_id = new_segment_model.id


@pytest.mark.vcr()
def test_get_segment():
    sw = get_sw()
    global segment_id
    print("segment_id = " + str(segment_id))
    model = sw.segment.get_by_id(segment_id)
    assert int(model.id) == segment_id


@pytest.mark.vcr()
def test_create_segment_incomplete_input():
    sw = get_sw()
    segment = Segment()
    with pytest.raises(SuperwiseValidationException):
        sw.segment.create(segment)
    with pytest.raises(SuperwiseValidationException):
        sw.segment.create(Segment(task_id=1, name="My Segment #3"))



import json
import os
import pathlib

import pandas as pd

from superwise.controller.summary.summary import get_summary_generator
from superwise.resources.superwise_enums import BooleanSecondaryType
from superwise.resources.superwise_enums import CategoricalSecondaryType
from superwise.resources.superwise_enums import FeatureType
from superwise.resources.superwise_enums import NumericSecondaryType
from tests import match_with_json_file


DATA_ENTITY_ASSETS_DIRECTORY = os.path.join(pathlib.Path(__file__).parent.absolute(), "resources/data_entity/assets")


def test_numeric_summary_generator():
    srs = pd.Series([i for i in range(0, 200)], name="my_numeric_feature")
    numeric_generator = get_summary_generator(FeatureType.NUMERIC.value)(srs)
    for secondary_type in NumericSecondaryType:
        f_name = os.path.join(DATA_ENTITY_ASSETS_DIRECTORY, "summary", "numeric", f"{secondary_type.name.lower()}.json")
        actual = json.loads(pd.Series(numeric_generator.generate_summary(secondary_type.value)).to_json())
        try:
            match_with_json_file(f_name, actual)
        except:
            raise Exception(f"test faild for {f_name}")


def test_categorical_summary_generator():
    print(CategoricalSecondaryType)
    for secondary_type in CategoricalSecondaryType:
        if secondary_type == CategoricalSecondaryType.CONSTANT:
            srs = pd.Series(["Category A", "Category A", "Category A"], name="my_categorical_feature")
            categorical_generator = get_summary_generator(FeatureType.CATEGORICAL.value)(srs)
        else:
            srs = pd.Series(["Category A", "Category B", "Category C"], name="my_categorical_feature")
            categorical_generator = get_summary_generator(FeatureType.CATEGORICAL.value)(srs)
        f_name = os.path.join(
            DATA_ENTITY_ASSETS_DIRECTORY, "summary", "categorical", f"{secondary_type.name.lower()}.json"
        )
        actual = json.loads(pd.Series(categorical_generator.generate_summary(secondary_type.value)).to_json())
        match_with_json_file(f_name, actual)


def test_boolean_summary_generator():
    for secondary_type in BooleanSecondaryType:
        if secondary_type == BooleanSecondaryType.FLAG:
            srs = pd.Series([True, False, None], name="my_boolean_feature").astype("category")
            srs = srs.cat.set_categories([True, False])
            boolean_generator = get_summary_generator(FeatureType.BOOLEAN.value)(srs)
        else:
            srs = pd.Series([0, 0, None], name="my_boolean_feature").astype("category")
            srs = srs.cat.set_categories([0, 1])
            boolean_generator = get_summary_generator(FeatureType.BOOLEAN.value)(srs)
        f_name = os.path.join(DATA_ENTITY_ASSETS_DIRECTORY, "summary", "boolean", f"{secondary_type.name.lower()}.json")
        actual = json.loads(pd.Series(boolean_generator.generate_summary(secondary_type.value)).to_json())
        match_with_json_file(f_name, actual)


def test_timestamp_summary_generator():
    srs = pd.Series([pd.Timestamp("01-01-2010"), pd.Timestamp("01-01-2010"), None], name="my_timestamp_feature")
    timestamp_generator = get_summary_generator(FeatureType.TIMESTAMP.value)(srs)
    f_name = os.path.join(DATA_ENTITY_ASSETS_DIRECTORY, "summary", f"timestamp.json")
    actual = timestamp_generator.generate_summary(None)
    match_with_json_file(f_name, actual)


def test_unknown_summary_generator():
    srs = pd.Series([None, None, None], name="my_unknown_feature")
    unknown_generator = get_summary_generator(FeatureType.UNKNOWN.value)(srs)
    f_name = os.path.join(DATA_ENTITY_ASSETS_DIRECTORY, "summary", f"unknown.json")
    actual = unknown_generator.generate_summary(None)
    match_with_json_file(f_name, actual)



""" This module implement roles functionality  """
from superwise.controller.base import BaseController
from superwise.models.task import Task


class RoleController(BaseController):
    """ Role controller class, implement functionalitiws for roles API """

    def __init__(self, client, sw):
        """
        constructer for RoleController class

        :param client:

        """
        super().__init__(client, sw)
        self.path = "model/v1/roles"
        self.model_name = "Role"

    def list_to_dict(self, model_list):
        """
        change list of models into dict

        :param model_list: list of models
        :return dict
        """
        dict = {}
        for model in model_list:
            properties = model.get_properties()
            dict[properties["description"]] = properties
        return dict



import json
import sys
from pprint import pprint

import pytest

from superwise import Superwise
from superwise.controller.exceptions import *
from superwise.models.task import Task
from superwise.resources.superwise_enums import TaskTypes
from tests import config
from tests import get_sw
from tests import print_results

task_id = None


@pytest.mark.vcr()
def test_create_task_inline():
    sw = get_sw()
    inline_model_test = sw.task.create(Task(name="name", description="this is description", monitor_delay=1))

    print_results("created task object 1", inline_model_test.get_properties())
    assert inline_model_test.name == "name"


@pytest.mark.vcr()
def test_create_task():
    sw = get_sw()
    model = Task()
    global task_id
    model.name = "this is test name"
    model.description = "description"
    model.monitor_delay = 0
    model.ongoing_label = 12
    model.name = "this is test name"
    new_task_model = sw.task.create(model)
    print_results("created task object 2", new_task_model.get_properties())
    assert new_task_model.name == "this is test name"
    assert new_task_model.description == "description"
    task_id = new_task_model.id


@pytest.mark.vcr()
def test_get_task():
    sw = get_sw()
    global task_id
    print(task_id)
    model = sw.task.get_by_id(task_id)
    assert int(model.id) == task_id


@pytest.mark.vcr()
def test_create_incomplete_input():
    sw = get_sw()
    model = Task()
    ok = False
    try:
        ok = True
        model = sw.task.create(model)
    except SuperwiseValidationException as e:
        assert ok == True
        pprint(e)
    print(model.get_properties())
    ok2 = False
    try:
        new_inline = sw.task.create(Task(description="inline tesk description", monitor_delay=1))
    except SuperwiseValidationException as e:
        pprint(e)
        ok2 = True
    assert ok2 == True



""" This module implement Task model  """
import enum

from superwise.models.base import BaseModel
from superwise.models.data_entity import DataEntity
from superwise.models.task import Task


class SegmentCondition(enum.Enum):
    __name__ = "SegmentCondition"
    GREATER_THAN_EQ = ">="
    LESS_THAN_EQ = "<="
    GREATER_THAN = ">"
    LESS_THAN = "<"
    IS_NULL = "is null"
    IN = "in"
    NOT_IN = "not in"
    EQUALS = "=="
    BETWEEN = "between"


class SegmentConditionDefinition:
    def __init__(self, entity_or_name, condition, value):
        self.entity_name = entity_or_name if not isinstance(entity_or_name, DataEntity) else entity_or_name.name
        self.condition = condition
        self.value = value
        assert isinstance(condition, SegmentCondition)

    def to_dict(self):
        return dict(entity_name=self.entity_name, condition=self.condition.value, value=self.value)


class Segment(BaseModel):
    """ Task model class, model  for tasks data """

    def __init__(
        self,
        id=None,
        task_or_id=None,
        name=None,
        status=None,
        definition_or_json=None,
        definition_query=None,
        created_at=None,
        created_by=None,
        archived_at=None,
        **kwargs
    ):
        """
        Constructor for Segment class

        :param task_or_id
        :param name
        :param definition
        :param created_by: (optional)
        """
        self.id = id
        if task_or_id is not None:
            self.task_id = task_or_id if not isinstance(task_or_id, Task) else task_or_id.id
        else:
            self.task_id = kwargs.get("task_id")
        self.name = name
        self.status = status
        if definition_or_json is not None:
            if any([isinstance(i, SegmentConditionDefinition) for i in definition_or_json]):
                self.definition_json = [i.to_dict() for i in definition_or_json]
            else:
                self.definition_json = definition_or_json
        else:
            self.definition_json = kwargs.get("definition_json")
        self.definition_query = definition_query
        self.created_at = created_at
        self.created_by = created_by
        self.archived_at = archived_at



import logging

from .categorical_summary_generator import CategoricalSummaryGenerator
from superwise.resources.superwise_enums import BooleanSecondaryType


logger = logging.getLogger(__name__)


class BooleanSummaryGenerator(CategoricalSummaryGenerator):
    def generate_summary(self, secondary_type: str):
        secondary_type_to_metrics = {
            BooleanSecondaryType.FLAG.value: [self.calc_missing_values, self.calc_new_values, self.calc_distribution],
            BooleanSecondaryType.NUMERIC.value: [
                self.calc_missing_values,
                self.calc_new_values,
                self.calc_distribution,
            ],
        }
        logger.debug(f"Generate summary for boolean ({secondary_type}) feature {self._entity.name}")
        summary = {}
        for metric in secondary_type_to_metrics[secondary_type]:
            summary = self._update(summary, metric())
        return summary



""" This module implement version functionality  """
import time

from superwise.config import Config
from superwise.controller.base import BaseController
from superwise.controller.exceptions import SuperwiseValidationException
from superwise.models.data_entity import DataEntity


class VersionController(BaseController):
    """ Version api controller """

    def __init__(self, client, sw):
        """
        Constructor of VersionController

        :param client: rest api client
        """
        super().__init__(client, sw)
        self.path = "model/v1/versions"
        self.model_name = "Version"

    def create(self, model, is_return_model=True, wait_until_complete=False, **kwargs):
        """

        :param model: model of version
        :return: model of version
        """
        if len(model.data_entities) and isinstance(model.data_entities[0], DataEntity):
            model.data_entities = [m.get_properties() for m in model.data_entities]
        res = super().create(model, **kwargs)
        if wait_until_complete:
            return self.get_by_id(res.id, wait_until_complete)
        else:
            return res

    def _dict(self, params, model_name=None):
        """

        :param params: dict
        :return: model of version
        """
        model_name = model_name or self.model_name

        model = super()._dict_to_model(params, model_name=model_name)

        if model_name != "DataEntity" and len(model.data_entities) and isinstance(model.data_entities[0], DataEntity):
            model.data_entities = [m.get_properties() for m in model.data_entities]
        return model

    def get_data_entities(self, version_id):
        """

        :param version_id:
        :return: list of models of data entities
        """
        models = []
        url = self.build_url("{}/{}/data_entities".format(self.path, version_id))
        self.logger.info("CALL GET DATA ENTITIES {} ".format(url))
        r = self.client.get(url)
        entities_lst = self.parse_response(r, "DataEntity", is_return_model=False)
        entities_lst = [e["data_entity"] for e in entities_lst]
        for entity in entities_lst:
            models.append(self._dict_to_model(entity, "DataEntity"))
        return models

    def get_by_id(self, idx, wait_until_complete=False):
        """ get by id implementation as we need to implement pooling version"""
        if wait_until_complete:
            while True:
                time.sleep(Config.POOLING_INTERVAL_SEC)
                model = super().get_by_id(idx)
                status = model.status
                if status in ["Pending"]:
                    self.logger.info(
                        "job status: Pending, summarizing should start soon, next pooling in %s sec",
                        Config.POOLING_INTERVAL_SEC,
                    )
                elif status == "Active":
                    self.logger.info("finished summarizing the version, return results")
                    return model
                elif status in ["Failed"]:
                    raise SuperwiseValidationException(model.get_properties())
                else:
                    raise Exception("unknown status %s", model.get_properties())
        else:
            return super().get_by_id(idx)

    def activate(self, version_id):
        url = "{}/{}".format(self.path, version_id)
        res = self.patch(url, {"status": "Active"})
        return res



""" This module implement Role model  """
from superwise.models.base import BaseModel


class Role(BaseModel):
    """ Role model class, model  for roles data """

    def __init__(
        self,
        id=None,
        task_type_id=None,
        is_nullable=None,
        description=None,
        internal_name=None,
        is_label=None,
        is_optional=None,
        role=None,
        value=None,
        **kwargs
    ):
        """
        constructer for Role class

        :param id:
        :param task_type_id:
        :param is_nullable:
        :param task_description:
        :param description:
        :param internal_name:
        :param is_label:
        :param is_optional:
        :param role:
        :param value:
        """
        self.task_type_id = task_type_id
        self.description = description
        self.internal_name = internal_name
        self.is_label = is_label
        self.is_optional = is_optional
        self.value = value
        self.is_nullable = is_nullable
        self.role = role
        self.id = id



import os
from multiprocessing import cpu_count
from multiprocessing import Pool

import numpy as np
import pandas as pd
import shap
from sklearn import preprocessing
from sklearn.ensemble import RandomForestRegressor
from sklearn.impute import SimpleImputer

from superwise import logger
from superwise.resources.superwise_enums import CategoricalSecondaryType
from superwise.resources.superwise_enums import FeatureType


class FeatureImportance:
    def __init__(self, entities_df):
        self._entities_df = entities_df
        self.logger = logger

    def _get_feature_mask(self):
        return self._entities_df[self._entities_df["role"] == "feature"]

    def get_numeric_features(self, exclude_metadata=False):
        feature_df = self._get_feature_mask()
        numeric_feature_df = feature_df[feature_df["type"] == FeatureType.NUMERIC.value]
        return numeric_feature_df["name"].tolist()

    def get_categorical_features(self, skip_sparse=True, exclude_metadata=False):
        feature_df = self._get_feature_mask()
        categorical_feature_df = feature_df[feature_df["type"] == FeatureType.CATEGORICAL.value]
        if skip_sparse:
            categorical_feature_df = categorical_feature_df[
                categorical_feature_df["secondary_type"] != CategoricalSecondaryType.SPARSE.value
            ]
        return categorical_feature_df["name"].tolist()

    def get_boolean_features(self, exclude_metadata=False):
        feature_df = self._get_feature_mask()
        boolean_feature_df = feature_df[feature_df["type"] == FeatureType.BOOLEAN.value]
        return boolean_feature_df["name"].tolist()

    def get_sparse_features(self, exclude_metadata=False):
        feature_df = self._get_feature_mask()

        categorical_feature_df = feature_df[feature_df["type"] == FeatureType.CATEGORICAL.value]
        sparse_feature_df = categorical_feature_df[
            categorical_feature_df["secondary_type"] == CategoricalSecondaryType.SPARSE.value
        ]
        return sparse_feature_df["name"].tolist()

    def _parallelize_shap(self, features, explainer, n_cores=cpu_count()):
        logger.debug("Running parallelize shap")
        features_split = np.array_split(features, n_cores)
        pool = Pool(n_cores)
        shap_values = np.concatenate(pool.map(explainer.shap_values, features_split))
        pool.close()
        pool.join()
        return shap_values

    def _create_categorical_embedding(self, srs):
        logger.debug("create categorical embedding for {}".format(srs.name))
        return srs.value_counts(dropna=False).to_dict()

    def _fi_pre_processing(self, baseline_data, version_entities) -> pd.DataFrame:
        logger.debug("pre processing for feature importance")
        features_for_modeling = (
            version_entities.get_boolean_features(exclude_metadata=True)
            + version_entities.get_numeric_features(exclude_metadata=True)
            + version_entities.get_categorical_features(exclude_metadata=True)
        )
        features_df = baseline_data.copy()[features_for_modeling]
        for feature in version_entities.get_categorical_features(exclude_metadata=True):
            logger.debug(f"feature importance pre pre-processing categorical feature {feature}")
            embedding = self._create_categorical_embedding(baseline_data[feature])
            features_df[feature] = pd.to_numeric(features_df[feature].map(embedding)).fillna(0)
        imp_mean = SimpleImputer(missing_values=np.nan, strategy="mean")
        numeric_features = version_entities.get_numeric_features(exclude_metadata=True)
        if len(numeric_features) > 0:
            logger.debug(f"pre-processing {len(numeric_features)} numeric features")
            features_df[numeric_features] = imp_mean.fit_transform(features_df[numeric_features])
        boolean_features = version_entities.get_boolean_features(exclude_metadata=True)
        if len(boolean_features) > 0:
            logger.debug(f"pre-processing {len(boolean_features)} boolean features")
            imp_mode = SimpleImputer(missing_values=np.nan, strategy="most_frequent")
            features_df[boolean_features] = features_df[boolean_features].astype(float)
            features_df[boolean_features] = imp_mode.fit_transform(features_df[boolean_features])
        return features_df

    def get_label_by_role(self, role):
        res = self._entities_df[self._entities_df["role"] == role]
        if len(res):
            return res.iloc[0]["name"]
        return None

    def _calc(self, baseline_data, target=None, sample=None):
        logger.debug("calc feature importance")
        if sample:
            baseline_data = baseline_data.sample(frac=sample)

        processed_features = self._fi_pre_processing(baseline_data, self)
        reconstruction_model = RandomForestRegressor(max_depth=7, n_estimators=50)
        label_name = None
        if target:
            label_name = target
        else:
            label_name = self.get_label_by_role("prediction value")
            if not label_name:
                label_name = self.get_label_by_role("label")
            if not label_name:
                raise Exception("no valid label to fit model")

        label = baseline_data[label_name]
        if not pd.api.types.is_numeric_dtype(label):
            le = preprocessing.LabelEncoder()
            le.fit(label)
            label = le.transform(label)
        reconstruction_model.fit(processed_features, label)
        explainer = shap.TreeExplainer(reconstruction_model)
        n_cores = int(os.environ.get("PARALLELIZE_CORES", cpu_count()))
        shap_values = self._parallelize_shap(processed_features, explainer, n_cores=n_cores)
        mean_shap = np.abs(shap_values).mean(axis=0)
        mean_shap = (mean_shap / mean_shap.sum()).round(4) * 100
        importance = dict(zip(processed_features.columns, mean_shap.tolist()))
        return importance

    def compute(self, data, target=None, sample=None):
        if "feature_importance" not in self._entities_df or pd.isnull(self._entities_df["feature_importance"]).all():
            logger.info("Feature importance was NOT provided as part of the version entities. Computing")
            feature_name_to_importance = self._calc(data, target, sample)
            self._entities_df["feature_importance"] = self._entities_df["name"].map(feature_name_to_importance)
        else:
            logger.info("Feature importance was provided as part of the version entities. Skipping Computation")
        self._entities_df["feature_importance"].fillna(0, inplace=True)
        return self._entities_df



import json
import os

import pytest

from superwise.models.task import Task
from superwise.models.version import Version
from tests import get_entities_fixture
from tests import get_sw
from tests import print_results


@pytest.mark.vcr()
def test_e2e_vcr():
    run_test_e2e()


def test_e2e_full():
    """
    No cassete, for real e2e tests
    """
    if os.environ.get("E2E_ENABLE"):
        run_test_e2e()
    else:
        print("E2E tests disables, skip")


def run_test_e2e():
    """
    1. task create
    2. version create
    version get (Wait)
    3. get dataentity
    3b. patch data entity
    4. activate the version
    """
    if not os.environ.get("E2E_ENABLE"):
        print("E2E tests disables, skip")
        return
    # create task
    sw = get_sw()
    inline_model_test = sw.task.create(
        Task(
            task_type_id=1,
            prediction=[
                {"value": 0, "is_default": 0, "description": "1"},
                {"value": 0, "is_default": 0, "description": "0"},
            ],
            title="inline title",
            task_description="inline tesk description",
            allow_prediction_update=True,
            fictive_label_mapper=[],
            monitor_delay=1,
            label=[
                {"value": 1, "is_default": 1, "description": "1"},
                {"value": 0, "is_default": 0, "description": "0"},
            ],
        )
    )

    print_results("created task object 1", inline_model_test.get_properties())
    assert inline_model_test.title == "inline title"
    task_id = inline_model_test.task_id

    # create segment
    """ FIXME KEY ERROR
    segment = sw.task.create_segment(
        task_id,
        name="Segment A",
        definition=[{"feature": "binary_str_null", "condition": ["in"], "value": ["0", "1"]}],
    )
    """

    ## create version
    entities = get_entities_fixture()

    for e in entities:
        print(e.get_properties())
    versionExternal = Version(
        task_id=task_id,
        version_name="test version",
        external_id=None,
        baseline_files=["gs://superwise-tools/integration_tests/basic/baseline_meta.parquet"],
        data_entities=entities,
    )

    model = sw.version.create(versionExternal)
    vid = model.id
    ### pooling results
    # model_after_polling = sw.version.get_by_id(vid)
    # print(model_after_polling)

    response = sw.version.activate(vid)
    # assert model.id == vid
    print(response.content)
    assert response.status_code == 204



""" This module implement data entities functionality  """
import pandas as pd

from superwise.config import Config
from superwise.controller.base import BaseController
from superwise.controller.exceptions import SuperwiseUnsupportedException
from superwise.controller.exceptions import SuperwiseValidationException
from superwise.controller.summary.entities_validator import EntitiesValidator
from superwise.controller.summary.feature_importance import FeatureImportance
from superwise.controller.summary.summary import Summary
from superwise.models.data_entity import DataEntity
from superwise.models.data_entity import DataEntitySummary
from superwise.resources.superwise_enums import CategoricalSecondaryType
from superwise.resources.superwise_enums import FeatureType


class DataEntityController(BaseController):
    """ controller for Data entities  """

    def __init__(self, client, sw):
        """
        constructer for DataEntityController class

        :param client:

        """
        super().__init__(client, sw)
        self.path = "model/v1/data_entities"
        self.model_name = "DataEntity"
        self._entities_df = None
        self.data = None

    @staticmethod
    def _pre_process_data(data):
        data.columns = data.columns.str.lower()
        for column in Config.LIST_DROP_DATA_COLS:
            if column in data.columns:
                data = data.drop(column, axis=1)
        return data

    def create(self, name=None, type=None, dimension_start_ts=None, role=None, feature_importance=None):
        """
        create data entity
        """

        params = locals()
        return self._dict_to_model(params)

    def update_summary(self, data_entity_id, summary):
        """
        update summary implementation
        """
        self.model = DataEntitySummary(data_entity_id, summary)
        self.model_name = "DataEntitySummary"
        self.create(self.model)

    def generate_summary(self, data_entities, task, data, base_version=None, **kwargs):
        """
        :param model: model of version
        :return: model of version
        """
        raise SuperwiseUnsupportedException(
            "Unsupported SDK API, please use  summarise() instead," " See CHANGELOG for more details"
        )

    def summarise(
        self,
        data,
        entities_dtypes,
        specific_roles,
        default_role="feature",
        importance_mapping=None,
        calc_importance=False,
        importance_target_label=None,
        importance_sample=None,
        base_version=None,
        **kwargs
    ):
        if importance_mapping and calc_importance:
            self.logger.debug("If importance dict is passed, calc shuld be false. override calc_importance")
            calc_importance = False
        if not entities_dtypes:
            raise Exception("not implemented yet")

        # self.data_entities = self.dataentities_from_dict(entities_dtypes)

        self.data = self._pre_process_data(data)

        data_entities = []
        for entity in entities_dtypes:
            importance_value = importance_mapping.get(entity, None) if importance_mapping else None
            data_entities.append(
                {
                    "name": entity,
                    "role": specific_roles.get(entity, default_role),
                    "type": entities_dtypes[entity],
                    "feature_importance": importance_value,
                    "dimension_start_ts": None,
                    "id": None,
                }
            )

        if base_version:
            if base_version.status not in ["Pending", "Active"]:
                raise SuperwiseValidationException(
                    "base version should be used for summarized only version, current status: {}".format(
                        base_version.status
                    )
                )
            r = self.client.get(self.build_url("model/v1/versions/{}/data_entities".format(base_version.id)))
            previus_entities = self.parse_response(r, is_return_model=False)
            previus_entities = [e["data_entity"] for e in previus_entities]
            previus_entities_dict = {}
            if previus_entities:
                for data_entity in previus_entities:
                    previus_entities_dict[data_entity["name"]] = data_entity["id"]
            for de in data_entities:
                if de["name"] in previus_entities_dict:
                    de["id"] = previus_entities_dict[de["name"]]
                else:
                    de["id"] = None

        entities_df = pd.DataFrame(data_entities)
        self._entities_df = entities_df
        validator = EntitiesValidator(entities_df, self.data)

        self.data = validator.prepare()
        if calc_importance:
            fi = FeatureImportance(self._entities_df)
            self._entities_df = fi.compute(self.data, target=importance_target_label, sample=importance_sample)
        entities_df_summary = Summary(self._entities_df, self.data).generate()
        data_entities = DataEntity.df_to_list(entities_df_summary)
        return data_entities



import json
import os
import pkgutil

try:
    config = json.loads(pkgutil.get_data("tests", "/conf/config.json"))
except Exception as ex:
    config = json.loads(pkgutil.get_data("tests", "/conf/config.example.json"))



import collections.abc
import logging
from typing import Dict

import pandas as pd


logger = logging.getLogger(__name__)


class EntitySummaryGenerator:
    def __init__(self, entity: pd.Series):
        self._entity = entity

    @classmethod
    def _update(cls, summary, measurements):
        for k, v in measurements.items():
            if isinstance(v, collections.abc.Mapping):
                summary[k] = cls._update(summary.get(k, {}), v)
            else:
                summary[k] = v
        return summary

    def calc_missing_values(self):
        logger.debug(f"Calc missing values for {self._entity.name}")
        missing_values = float(self._entity.isna().sum() / len(self._entity))
        return {"statistics": {"missing_values": missing_values}}

    def generate_summary(self, secondary_type) -> Dict:
        logger.debug(f"Calc generic summary for {self._entity.name}")
        return self.calc_missing_values()



import json
import sys
import uuid
from pprint import pprint

import boto3
import pytest
import requests
from google.cloud import storage
from google.oauth2 import service_account
from requests import Response

from project_root import PROJECT_ROOT
from superwise import Client
from superwise import Superwise
from superwise.controller.exceptions import *
from superwise.models.task import Task
from tests import config
from tests import get_sw
from tests import print_results


@pytest.fixture(scope="function")
def mock_get_token(monkeypatch):
    monkeypatch.setattr(
        Client,
        "get_token",
        lambda *args, **kwargs: "eyJhbGciOiJSUzI1NiIsInR5cCI6IkpXVCIsImtpZCI6IjQ3ZDdmMDg2In0.eyJzdWIiOiI5YzNlZmUxZC03NGNlLTRlZTItYTMyOC1kMWZmNmQyMDAyM2YiLCJlbWFpbCI6InN3X2JhcmFrQHN1cGVyd2lzZS5haSIsInVzZXJNZXRhZGF0YSI6e30sInRlbmFudElkIjoiYmFyYWsiLCJyb2xlcyI6WyJWaWV3ZXIiXSwicGVybWlzc2lvbnMiOlsiZmUuc2VjdXJlLndyaXRlLnVzZXJBcGlUb2tlbnMiLCJmZS5zZWN1cmUuZGVsZXRlLnVzZXJBcGlUb2tlbnMiLCJmZS5zZWN1cmUucmVhZC51c2VyQXBpVG9rZW5zIl0sIm1ldGFkYXRhIjp7fSwiY3JlYXRlZEJ5VXNlcklkIjoiNDg5ZmM5Y2YtZDlhYy00MWMwLWJmM2ItN2VhNDUyNDY4ODEyIiwidHlwZSI6InVzZXJBcGlUb2tlbiIsInVzZXJJZCI6IjQ4OWZjOWNmLWQ5YWMtNDFjMC1iZjNiLTdlYTQ1MjQ2ODgxMiIsImlhdCI6MTYzNjY0ODIyMywiZXhwIjoxNjM2NzM0NjIzLCJpc3MiOiJmcm9udGVnZyJ9.qhEclIsSpfwXpCTFb8qhKpizRWtpQSnkE7VMsy9Et3guLcOcTiTVZ2wOJPmemtL3g3AStKH2jFSOEwQOoqnvgSR3dum9I_Ae3UwrFNRnM3EqOz7UsD0cJAd1AYy-69-67o5oX9A2U4MPZSA5Dr5Edbvn86-AsBJhADGDs5AyEyuGmlJTq0ACGAmoC8qZlxwnOsn9wIzTiQVU7085M73n5iJ26SNhsy4KNpU-8oR2lC1akDroHzL8aIr5dAWSWZz_cfcyWQyC1gqb4_ZAvG1GXiKwsGW2irFyfGoD9zrwMoMGuWXKCbXnHxIzuv8ImX_cRVPXq5xVBYUXwODr83Q3FA",
    )
    monkeypatch.setattr(Client, "get_service_account", lambda *args, **kwargs: {})


@pytest.fixture(scope="function")
def sw(mock_gcp_client):
    return Superwise(client_id="test", secret="test")


@pytest.fixture(scope="function")
def mock_transaction_requests(monkeypatch):
    the_response = Response()
    the_response._content = b'{ "transaction_id" : "123" }'
    the_response.status_code = 201
    monkeypatch.setattr(requests, "post", lambda *args, **kwargs: the_response)


@pytest.fixture(scope="function")
def mock_boto_client(monkeypatch):
    class BotoClient:
        def __init__(self, *args, **kwargs):
            pass

        def get_object(self, *args, **kwargs):
            class ObjectResposne:
                def read(self):
                    return "blablabla"

            return dict(Body=ObjectResposne())

    monkeypatch.setattr(boto3, "client", lambda *args, **kwargs: BotoClient())


@pytest.fixture(scope="function")
def mock_gcp_client(monkeypatch):
    class GCSClient:
        def __init__(self, *args, **kwargs):
            self.name = "test"

        def bucket(self, bucket_name):
            return GCSClient()

        def blob(self, file_name):
            return GCSClient()

        def download_as_string(self):
            return "asdasdaas"

        def upload_from_string(self, data):
            return None

    monkeypatch.setattr(service_account.Credentials, "from_service_account_info", lambda *args, **kwargs: "")
    monkeypatch.setattr(storage, "Client", lambda *args, **kwargs: GCSClient())


@pytest.fixture(scope="function")
def get_transaction_records_payload():
    with open(f"{PROJECT_ROOT}/tests/resources/transaction/records_payload.json") as f:
        return json.loads(f.read())


def test_transaction_records(mock_transaction_requests, mock_get_token, sw, get_transaction_records_payload):
    status = sw.transaction.log_batch(task_id=1, records=get_transaction_records_payload)
    assert isinstance(status, dict) and "transaction_id" in status.keys()
    status = sw.transaction.log_batch(task_id="test", version_id="test", records=get_transaction_records_payload)
    assert isinstance(status, dict) and "transaction_id" in status.keys()


def test_transaction_file(mock_transaction_requests, mock_get_token, sw):
    status = sw.transaction.log_file("gs://fvsdfvfdv")
    assert isinstance(status, dict) and "transaction_id" in status.keys()


def test_transaction_with_wrong_file_path(mock_transaction_requests, mock_get_token, sw):
    ok = False
    try:
        status = sw.transaction.log_file("wrong path")
    except Exception as e:
        assert str(e) == "transaction file failed because of wrong file path. file path should be gcs or s3 path."
        ok = True
    assert ok is True


def test_transaction_upload_from_gcs(mock_transaction_requests, mock_get_token, mock_gcp_client, sw):
    status = sw.transaction.log_from_gcs(
        file_path="gs://superwise-oryan-test/new_integration_tests_binary_classification_predictions.csv",
        service_account={},
    )

    assert isinstance(status, dict) and "transaction_id" in status.keys()


def test_transaction_upload_from_s3(mock_transaction_requests, mock_get_token, mock_gcp_client, mock_boto_client, sw):
    status = sw.transaction.log_from_s3(
        file_path="s3://superwise-oryan-test/new_integration_tests_binary_classification_predictions.csv",
        aws_access_key_id="test",
        aws_secret_access_key="test",
    )
    assert isinstance(status, dict) and "transaction_id" in status.keys()



""" This module implement Task model  """
from superwise.models.base import BaseModel


class Task(BaseModel):
    """ Task model class, model  for tasks data """

    def __init__(
        self,
        id=None,
        external_id=None,
        name=None,
        description=None,
        monitor_delay=None,
        time_units=None,
        is_archive=None,
        **kwargs
    ):
        """
        constructer for Task class
        :param id:
        :param external_id:
        :param name:
        :param description:
        :param monitor_delay:
        :param time_units:
        :param is_archive:
        """
        self.external_id = external_id
        self.name = name
        self.id = id or None
        self.description = description
        self.monitor_delay = monitor_delay
        self.time_units = time_units
        self.is_archive = is_archive



""" This module implement tasks functionality  """
from superwise.controller.base import BaseController


class TaskController(BaseController):
    """Class TaskController - responsible for task functionality"""

    def __init__(self, client, sw):
        """
        Constructor of TaskController class
        """
        super().__init__(client, sw)
        self.path = "admin/v1/tasks"
        self.model_name = "Task"



import logging
import os

import numpy as np
import pandas as pd
import scipy.stats

from superwise.resources.superwise_enums import BooleanSecondaryType
from superwise.resources.superwise_enums import CategoricalSecondaryType
from superwise.resources.superwise_enums import DataTypesRoles
from superwise.resources.superwise_enums import FeatureType
from superwise.resources.superwise_enums import NumericSecondaryType
from superwise.resources.superwise_enums import TaskTypes

logger = logging.getLogger(__name__)


def compute_secondary_numeric(skewness):
    if skewness > 2:
        return NumericSecondaryType.NUM_RIGHT_TAIL.value
    if skewness > -2:
        return NumericSecondaryType.NUM_CENTERED.value
    return NumericSecondaryType.NUM_LEFT_TAIL.value


def compute_secondary_categorical(row):
    if row["num_unique"] == 1:
        return CategoricalSecondaryType.CONSTANT.value
    if row["num_unique"] < 250:
        return CategoricalSecondaryType.DENSE.value
    return CategoricalSecondaryType.SPARSE.value


def compute_secondary_boolean(row):
    if pd.api.types.is_numeric_dtype(row["pandas_type"]):
        return BooleanSecondaryType.NUMERIC.value
    else:
        return BooleanSecondaryType.FLAG.value


class EntitiesValidationError(RuntimeError):
    def __init__(self, reason):
        super(EntitiesValidationError, self).__init__(reason)
        self.reason = reason


class EntitiesValidator:
    def __init__(self, version_entities_df, data, MAX_DIMENSIONS=None):
        """
            version_entities_df
                name | type | role | is_dimension
            task
                task_type_id
                label -> mapper of external to internal
                prediction -> mapper of external to internal
        """
        self.MAX_DIMENSIONS = MAX_DIMENSIONS or int(os.environ.get("MAX_DIMENSIONS", 10))
        self._version_entities_df = version_entities_df
        self._data = data

    def _assert_type(self, column, column_description, expected_type):
        column_version_entities = self._version_entities_df[self._version_entities_df["name"] == column].iloc[0]
        if column_version_entities["type"] != expected_type:
            error_msg = (
                f"Column {column} ({column_description}) has type {column_version_entities['type']} "
                f"but type {expected_type} is required"
            )
            raise EntitiesValidationError(error_msg)

    @staticmethod
    def assert_range(data, column, column_description, expected_min, expected_max):
        minimum = data[column["name"]].min()
        maximum = data[column["name"]].max()
        if minimum < expected_min or maximum > expected_max:
            error_msg = (
                f"column={column['name']} ({column_description}) has values ranging from "
                f"{minimum} to {maximum} whereas {expected_min} to {expected_max} is expected"
            )
            raise EntitiesValidationError(error_msg)

    @staticmethod
    def _assert_not_null(data, column, column_description):
        if data[column].isna().any():
            error_msg = "Column {} ({}) has null values".format(column, column_description)
            raise EntitiesValidationError(error_msg)

    @staticmethod
    def _assert_not_empty(data, column, column_description):
        if data[column].isna().all():
            error_msg = f"Column {column} ({column_description}) has all null values"
            raise EntitiesValidationError(error_msg)

    @staticmethod
    def _assert_subset(set_a, set_a_description, set_b, set_b_description):
        """
            asserts a is a subset of b
        """
        diff = set_a - set_b
        if diff:
            error_msg = (
                f"These columns exist in {set_a_description} but are missing from {set_b_description}: "
                f"{','.join(map(str, diff))}"
            )
            raise EntitiesValidationError(error_msg)

    def _is_valid_record_id(self):
        id_column = self._version_entities_df[self._version_entities_df["role"] == "id"]
        if len(id_column) > 1:
            raise Exception("more then one entity with role = id")
        elif len(id_column) == 0:
            raise Exception("Record id col is missing")

        id_column = id_column.iloc[0]["name"]
        if not self._data[id_column].is_unique:
            error_msg = f"Column {id_column} is marked as role ID but has duplicated values"
            raise EntitiesValidationError(error_msg)

    def _is_valid_label_weight(self):
        label_weight_col = self._version_entities_df["role"] == DataTypesRoles.LABEL_WEIGHT.value
        label_weight_size = label_weight_col.sum()
        if label_weight_size > 1:
            raise EntitiesValidationError("more then one ({}) label weight entity".format(label_weight_size))
        elif label_weight_size:
            label_weight_df = self._version_entities_df[label_weight_col].iloc[0]["name"]
            self._assert_type(label_weight_df, "label weight", FeatureType.NUMERIC.value)
            self._data["label_weight"] = self._data["label_weight"].fillna(1).astype(float)
        else:
            self._data["label_weight"] = 1

    def _version_entities_has_not_reserved_names(self):
        reserved_names = ["task_id", "version_id"]
        reserved_words_used = self._version_entities_df[self._version_entities_df["name"].isin(reserved_names)][
            "name"
        ].tolist()
        if reserved_words_used:
            error_msg = f"version_entities contains reserved words {','.join(reserved_words_used)}"
            raise EntitiesValidationError(error_msg)

    def _has_all_required_roles(self):
        existing_role = set(self._version_entities_df["role"].unique())
        required_roles = set([DataTypesRoles.ID.value, DataTypesRoles.TIMESTAMP.value])
        self._assert_subset(required_roles, "required roles", existing_role, "existing roles")

    def _column_names_match_version_entities(self):
        data_feature_names = set(self._data.columns)
        version_entities_feature_names = set(self._version_entities_df["name"].tolist())
        self._assert_subset(data_feature_names, "data", version_entities_feature_names, "version_entities")
        self._assert_subset(version_entities_feature_names, "version_entities", data_feature_names, "data")

    def _is_valid_dimension_ts_count(self):
        df = self._version_entities_df
        cnt = len(df[~df["dimension_start_ts"].isnull()])
        if cnt > self.MAX_DIMENSIONS:
            error_msg = f"Maximum allowed dimensions is {self.MAX_DIMENSIONS}. You requested {cnt}"
            raise EntitiesValidationError(error_msg)

    def _is_valid_dimension(self):
        dimensions = self._version_entities_df[~self._version_entities_df["dimension_start_ts"].isnull()][
            "name"
        ].tolist()
        for dimension in dimensions:
            self._assert_type(dimension, "dimension", FeatureType.CATEGORICAL.value)

    def _is_valid_prediction_probability(self):
        columns = self._version_entities_df[self._version_entities_df["role"] == "prediction probability"]
        for idx, column in columns.iterrows():
            self._assert_type(column["name"], "prediction probability", FeatureType.NUMERIC.value)
            self._assert_not_null(self._data, column["name"], "prediction_probability")
            self.assert_range(self._data, column, "prediction probability", 0, 1)

    def _is_valid_prediction_ts(self):
        column = self._version_entities_df[self._version_entities_df["role"] == "time stamp"]
        if len(column) > 1:
            raise EntitiesValidationError("more then one entity with role 'time stamp'")
        column = column.iloc[0]
        self._assert_type(column["name"], "prediction time stamp", FeatureType.TIMESTAMP.value)
        self._assert_not_null(self._data, column["name"], "prediction time stamp")

    def _is_valid_label_ts(self):
        column = self._version_entities_df[self._version_entities_df["role"] == "label time stamp"]
        if len(column) > 1:
            raise EntitiesValidationError("only one 'label time stamp' entity allowed")
        elif len(column) == 0:
            return True
        column = column.iloc[0]
        self._assert_type(column["name"], "label time stamp", FeatureType.TIMESTAMP.value)
        self._assert_not_null(self._data, column["name"], "label time stamp")

    def _time_stamp_types_valid(self):
        ts_columns = self._version_entities_df[self._version_entities_df["type"] == FeatureType.TIMESTAMP.value][
            "name"
        ].tolist()
        for ts_column in ts_columns:
            self._assert_not_empty(self._data, ts_column, "time stamp")
            try:
                self._data[ts_column] = pd.to_datetime(self._data[ts_column])
            except Exception:
                error_msg = "Column {} is marked as type Timestamp but is malformed".format(ts_column)
                raise EntitiesValidationError(error_msg)

    def _boolean_types_valid(self):
        boolean_columns = self._version_entities_df[self._version_entities_df["type"] == FeatureType.BOOLEAN.value][
            "name"
        ].tolist()
        for boolean_column in boolean_columns:
            self._assert_not_empty(self._data, boolean_column, "boolean")

            unique_values = set(self._data[boolean_column].dropna().unique())
            if len(unique_values) == 0:
                error_msg = "Boolean column {} is empty".format(boolean_column)
                raise EntitiesValidationError(error_msg)

            if not (unique_values <= {True, False} or unique_values <= {1, 0}):
                error_msg = (
                    "Boolean column must be either a subset of {{True, False}} or {{1, 0}}. "
                    "You provided {} for column {}".format(",".join(map(str, unique_values)), boolean_column)
                )
                raise EntitiesValidationError(error_msg)
            self._data[boolean_column] = self._data[boolean_column].astype("boolean")

    def _numeric_types_valid(self):
        numeric_columns = self._version_entities_df[self._version_entities_df["type"] == FeatureType.NUMERIC.value][
            "name"
        ].tolist()
        for numeric_column in numeric_columns:
            self._assert_not_empty(self._data, numeric_column, "numeric")
            self._data[numeric_column] = self._data[numeric_column].astype("float")

    def _categorical_types_valid(self):
        categorical_columns = self._version_entities_df[
            self._version_entities_df["type"] == FeatureType.CATEGORICAL.value
        ]["name"].tolist()
        for categorical_column in categorical_columns:
            self._assert_not_empty(self._data, categorical_column, "categorical")
            self._data[categorical_column] = self._data[categorical_column].astype("category")

    def _build_feature_statistics_df(self, data):
        types = []
        for feature_name in data.columns:
            feature_data = data[feature_name].dropna()
            feature_stat = dict()
            feature_stat["name"] = feature_name
            feature_stat["pandas_type"] = data[feature_name].dtype
            hist = feature_data.value_counts(normalize=True)
            feature_stat["num_unique"] = len(hist)
            feature_stat["% unique"] = len(hist) / len(data)
            feature_stat["% top 5"] = hist.head(5).sum()
            feature_stat["skewness"] = (
                feature_data.skew() if pd.api.types.is_numeric_dtype(feature_stat["pandas_type"]) else np.nan
            )
            types.append(feature_stat)
        types = pd.DataFrame(types)
        types["type"] = types["name"].map(self._version_entities_df.set_index("name")["type"])
        return types

    @staticmethod
    def _infer_secondary_type_row(row):
        if row["type"] == FeatureType.CATEGORICAL.value:
            return compute_secondary_categorical(row)
        if row["type"] == FeatureType.BOOLEAN.value:
            return compute_secondary_boolean(row)
        if row["type"] == FeatureType.NUMERIC.value:
            return compute_secondary_numeric(row["skewness"])
        return np.nan

    def infer_secondary_type(self):
        if (
            "secondary_type" not in self._version_entities_df
            or self._version_entities_df["secondary_type"].isnull().values.all()
        ):
            logger.info("Secondary type was NOT provided as part of the version entities. Computing")
            feature_statistics_df = self._build_feature_statistics_df(self._data)
            feature_statistics_df["secondary_type"] = feature_statistics_df.apply(
                self._infer_secondary_type_row, axis=1
            )
            self._version_entities_df["secondary_type"] = self._version_entities_df["name"].map(
                feature_statistics_df.set_index("name")["secondary_type"]
            )
        elif self._version_entities_df["secondary_type"].isnull().values.any():
            raise EntitiesValidationError("seconday_type supplied for some entities, should be supplied for all or non")

    def infer_data_type(self):
        logger.info("Infer parquet type based on type")
        logger.debug("Type map -> Unknown=string, numeric=float64, timestamp=timestamp, boolean=bool_")
        data_types = {
            FeatureType.UNKNOWN.value: "text",
            FeatureType.NUMERIC.value: "number",
            FeatureType.TIMESTAMP.value: "timestamp",
            FeatureType.BOOLEAN.value: "bool",
        }
        self._version_entities_df["data_type"] = self._version_entities_df["type"].map(data_types)
        logger.info("Infer parquet type for categorical based on data")
        for feature_name in self._version_entities_df.loc[
            self._version_entities_df["type"] == FeatureType.CATEGORICAL.value, "name"
        ].tolist():
            try:
                self._data[feature_name].astype("float")
                self._version_entities_df.loc[self._version_entities_df["name"] == feature_name, "data_type"] = "number"
                logger.debug(f"Categorical feature {feature_name} has parquet type float64")
            except:
                self._version_entities_df.loc[self._version_entities_df["name"] == feature_name, "data_type"] = "text"
                logger.debug(f"Categorical feature {feature_name} has parquet type string")

    def convert_to_data_type(self):
        logger.info("Convert to inferred data type")
        data_type_to_conversion = {
            "text": lambda col: col.astype("string"),
            "number": lambda col: col.astype("float"),
            "bool": lambda col: col.astype("boolean"),
            "timestamp": lambda col: pd.to_datetime(col, format="%Y-%m-%d %H:%M:%S"),
        }
        for version_entity in self._version_entities_df.to_dict(orient="records"):
            feature_name = version_entity["name"]
            data_type = version_entity["data_type"]
            self._data[feature_name] = data_type_to_conversion[data_type](self._data[feature_name])
        logger.info("Conversion to inferred data type finished")

    def prepare(self):
        """
        Validate and prepare the data
        """
        logger.info("start validate and prepare data")
        self._data = self._data.reset_index(drop=True)

        # version_entities attributes
        self._version_entities_has_not_reserved_names()
        self._column_names_match_version_entities()

        logger.info("start validate and prepare dimensions")
        # Dimensions
        self._is_valid_dimension_ts_count()
        self._is_valid_dimension()

        logger.info("start validate and prepare types")
        # Types
        self._boolean_types_valid()
        self._categorical_types_valid()
        self._time_stamp_types_valid()
        self._numeric_types_valid()

        logger.info("start validate and prepare roles")
        # Roles
        self._has_all_required_roles()
        self._is_valid_record_id()
        self._is_valid_prediction_ts()
        self._is_valid_prediction_probability()
        self._is_valid_label_ts()
        self._is_valid_label_weight()
        self.infer_secondary_type()
        self.infer_data_type()
        self.convert_to_data_type()
        return self._data



""" This module implement DataEntity model  """
import json

import pandas as pd

from superwise.models.base import BaseModel


class DataEntity(BaseModel):
    """ data entity model class """

    def __init__(
        self,
        id=None,
        dimension_start_ts=None,
        type=None,
        name=None,
        role=None,
        feature_importance=None,
        summary=None,
        secondary_type=None,
        data_type=None,
        **kwargs
    ):
        """
        constructer for DataEntity class

        :param type:
        :param name:
        :param is_dimension:
        :param role:
        :param feature_importance:
        """
        self.name = name.lower() if name else None
        self.type = self.get_enum_value(type)
        self.role = self.get_enum_value(role)
        self.feature_importance = feature_importance
        self.summary = summary
        self.secondary_type = secondary_type
        self.id = id
        self.dimension_start_ts = self.from_datetime(dimension_start_ts)
        self.data_type = data_type

    @staticmethod
    def list_to_df(data_entities):
        data = [d.get_properties() for d in data_entities]
        df = pd.DataFrame(data)
        return df

    @staticmethod
    def df_to_list(data_entities_df):
        data = json.loads(data_entities_df.to_json(orient="records"))
        entities = []
        [entities.append(DataEntity(**d)) for d in data]
        return entities


class DataEntitySummary(BaseModel):
    """ summary model class """

    def __init__(self, idx=None, summary=None):
        """
        constructer for DataEntitySummary class

        :param idx:
        :param summary:
        """
        self.id = idx
        self.summary = summary



""" This module implement data custom exceptions  """


class SuperWiseValidationError(Exception):
    """  SuperWiseValidationError customn exception """


class SuperwiseException(SuperWiseValidationError):
    """  SuperwiseException customn exception """


class SuperwiseTokenException(SuperWiseValidationError):
    """  SuperwiseException customn exception """


class SuperwiseValidationException(SuperWiseValidationError):
    """  SuperwiseValidationException customn exception """


class SuperwiseUnsupportedException(SuperWiseValidationError):
    """  SuperwiseValidationException customn exception """


class SuperwiseAuthException(SuperWiseValidationError):
    """  SuperwiseAuthException customn exception """


class SuperwiseServiceAccountException(SuperWiseValidationError):
    """  SuperwiseAuthException custom exception """


class SuperwiseStorageDownloadS3Error(SuperWiseValidationError):
    """  SuperwiseStorageDownloadS3Error custom exception """


class SuperwiseStorageDownloadGCSError(SuperWiseValidationError):
    """  SuperwiseStorageDownloadGCSError custom exception """


class SuperwiseStorageUploadGCSError(SuperWiseValidationError):
    """  SuperwiseStorageUploadGCSError custom exception """



import json
import sys
from pprint import pprint

import pytest

from superwise import Superwise
from superwise.controller.exceptions import *
from superwise.models.task import Task
from tests import config
from tests import get_sw
from tests import print_results


@pytest.mark.vcr()
def test_task_type_roles():
    ## FIXME - for gui bring back later
    return
    sw = get_sw()
    model = sw.role.get_by_task_type_id(1)
    for m in model:
        print(m.get_properties())
    assert len(model) == 9



import logging

import numpy as np
import pandas as pd
from sklearn.linear_model import LinearRegression
from sklearn.preprocessing import KBinsDiscretizer

from .entity_summary_generator import EntitySummaryGenerator
from superwise.resources.superwise_enums import NumericSecondaryType

logger = logging.getLogger(__name__)


class NumericalSummaryGenerator(EntitySummaryGenerator):
    def __init__(self, entity):
        super(NumericalSummaryGenerator, self).__init__(entity)

    def _get_linear_percentile(self, quantile_deviation=0.03):
        # upper limit
        q_values = self._entity.astype(float).quantile(q=[0.98, 0.985, 0.99, 0.995])
        lr = LinearRegression().fit(q_values.reset_index()[["index"]], q_values.values)
        q_predict = 1 + quantile_deviation
        upper_limit = float(lr.predict(np.array(q_predict).reshape(-1, 1))[0])
        # Lower limit
        q_values = self._entity.astype(float).quantile(q=[0.005, 0.01, 0.015, 0.02])
        lr = LinearRegression().fit(q_values.reset_index()[["index"]], q_values.values)
        q_predict = 0 - quantile_deviation
        lower_limit = float(lr.predict(np.array(q_predict).reshape(-1, 1))[0])
        return lower_limit, upper_limit

    def calc_range(self):
        logger.debug("Calc range for".format(self._entity.name))
        lower_limit, upper_limit = self._get_linear_percentile()
        entities_count = self._entity.notna().sum()
        if entities_count:
            is_outlier_mask = (lower_limit > self._entity) | (self._entity > upper_limit)
            outliers_count = (is_outlier_mask & self._entity.notna()).sum()
            outliers = outliers_count / entities_count
        else:
            outliers = 0
        return {"range": {"from": lower_limit, "to": upper_limit}, "statistics": {"outliers": outliers}}

    def calc_min_value(self):
        logger.debug(f"Calc min value for {self._entity.name}")
        min_value = self._entity.min()
        min_value = None if pd.isna(min_value) else float(min_value)
        return {"statistics": {"min": min_value}}

    def calc_max_value(self):
        logger.debug(f"Calc max value for {self._entity.name}")
        max_value = self._entity.max()
        max_value = None if pd.isna(max_value) else float(max_value)
        return {"statistics": {"max": max_value}}

    def calc_std(self):
        logger.debug(f"Calc std value for {self._entity.name}")
        stdev = self._entity.std(ddof=1)
        logger.debug("calculated std {}".format(stdev))
        stdev = None if pd.isna(stdev) else float(stdev)
        return {"statistics": {"std": stdev}}

    def calc_mean_value(self):
        logger.debug(f"Calc mean value for {self._entity.name}")
        mean_value = self._entity.mean()
        mean_value = None if pd.isna(mean_value) else float(mean_value)
        return {"statistics": {"mean": mean_value}}

    def _numeric_bins(self, n_bins=30, strategy="kmeans", pad_bins=2):
        n_bins = min(self._entity.nunique(), n_bins)
        discretizer = KBinsDiscretizer(n_bins=n_bins, encode="ordinal", strategy=strategy)
        bins = discretizer.fit(self._entity.dropna().to_frame()).bin_edges_[0]

        if pad_bins:
            lower_bin_width = bins[1] - bins[0]
            lower_bins_padding = [bins[-1] + lower_bin_width * (i + 1) for i in range(pad_bins)]
            upper_bin_width = bins[-1] - bins[-2]
            upper_bins_padding = [bins[0] - upper_bin_width * (i + 1) for i in range(pad_bins)]
            bins = np.append(bins, lower_bins_padding + upper_bins_padding)
            bins.sort()
        return bins

    def calc_numeric_distribution(self):
        logger.debug(f"Calc numeric distribution for {self._entity.name}")
        bins = self._numeric_bins()
        return self._calc_distribution(bins)

    def _calc_distribution(self, bins):
        bins = np.append(bins, [-np.Inf, np.Inf])
        bins.sort()
        buckets = pd.cut(self._entity, bins, right=False)
        buckets = buckets.value_counts(normalize=True, dropna=True).sort_index()
        buckets = pd.DataFrame(buckets).reset_index().reset_index().sort_values(by="index")
        buckets.columns = ["id", "name", "frequency"]
        buckets["lower"] = buckets["name"].values.categories.left
        buckets["upper"] = buckets["name"].values.categories.right

        mid_bin = list()
        bin_width = list()
        for idx, row in buckets.dropna().iterrows():
            if not {row["name"].right, row["name"].left}.intersection({np.Inf, -np.Inf}):
                mid_bin.append(row["name"].mid)
                bin_width.append(row["name"].length)

        if len(bin_width) == 0:
            bins_without_inf = bins[(bins < bins.max()) & (bins > bins.min())]
            mid_bin = [bins_without_inf[0] - 1, bins_without_inf[0] + 1]
            bin_width = [1, 1]
        else:
            bin_width.insert(0, bin_width[0])
            mid_bin.insert(0, buckets.iloc[0]["name"].right - bin_width[0] / 2)
            bin_width.append(bin_width[-1])
            mid_bin.append(buckets.dropna().iloc[-1]["name"].left + bin_width[-1] / 2)

        buckets["name"] = buckets["name"].astype(str)
        buckets["mid_bin"] = mid_bin
        buckets["bin_width"] = bin_width

        # remove -Inf & Inf
        bins = bins[1:-1]
        return {"distribution": {"bins": bins, "buckets": buckets.to_dict(orient="records")}}

    def generate_summary(self, secondary_type):
        secondary_type_to_metrics = {
            NumericSecondaryType.NUM_RIGHT_TAIL.value: [
                self.calc_missing_values,
                self.calc_range,
                self.calc_min_value,
                self.calc_max_value,
                self.calc_mean_value,
                self.calc_numeric_distribution,
                self.calc_std,
            ],
            NumericSecondaryType.NUM_LEFT_TAIL.value: [
                self.calc_missing_values,
                self.calc_range,
                self.calc_min_value,
                self.calc_max_value,
                self.calc_mean_value,
                self.calc_std,
                self.calc_numeric_distribution,
            ],
            NumericSecondaryType.NUM_CENTERED.value: [
                self.calc_missing_values,
                self.calc_range,
                self.calc_min_value,
                self.calc_max_value,
                self.calc_mean_value,
                self.calc_std,
                self.calc_numeric_distribution,
            ],
        }
        logger.debug(f"Generate summary for Numerical ({secondary_type}) feature {self._entity.name}")
        summary = {}
        for metric in secondary_type_to_metrics[secondary_type]:
            summary = self._update(summary, metric())
        return summary



""" This module implement version  model  """
import pandas as pd

from superwise.models.base import BaseModel


class Version(BaseModel):
    """ Version model class, model  for version data """

    def __init__(
        self,
        id=None,
        task_id=None,
        name=None,
        created_at=None,
        data_entities=None,
        baseline_df=None,
        status=None,
        **kwargs
    ):
        """
        constructor for Version class

        :param task_id:
        :param name:
        :param client_name:
        :param external_id:
        :param created_at:
        :param data_entities:
        """
        self.id = id
        self.task_id = task_id
        self.name = name
        self.baseline_files = []
        self.created_at = created_at
        self.data_entities = data_entities or []
        self.status = status
        self.baseline_df = baseline_df



from enum import Enum


class DataTypesRoles(Enum):
    """ Enum of Roles"""

    ID = "id"
    TIMESTAMP = "time stamp"
    FEATURE = "feature"
    PREDICTION_PROBABILITY = "prediction probability"
    PREDICTION_VALUE = "prediction value"
    LABEL = "label"
    LABEL_TIMESTAMP = "label time stamp"
    LABEL_WEIGHT = "label weight"
    METADATA = "metadata"


class TaskTypes(Enum):
    """ Enum of DataTypes"""

    BINARY_CLASSIFICATION = "Binary Classification"
    BINARY_ESTIMATION = "Binary Estimation"
    REGRESSION = "Regression"
    MULTICLASS_CLASSIFICATION = "Multiclass Classification"


class FeatureType(Enum):
    """ Enum of FeatureType"""

    NUMERIC = "Numeric"
    BOOLEAN = "Boolean"
    CATEGORICAL = "Categorical"
    TIMESTAMP = "Timestamp"
    UNKNOWN = "Unknown"


class CategoricalSecondaryType(Enum):
    CONSTANT = "Cat_constant"
    DENSE = "Cat_dense"
    SPARSE = "Cat_sparse"


class NumericSecondaryType(Enum):
    NUM_RIGHT_TAIL = "Num_right_tail"
    NUM_LEFT_TAIL = "Num_left_tail"
    NUM_CENTERED = "Num_centered"


class BooleanSecondaryType(Enum):
    FLAG = "Boolean_flag"
    NUMERIC = "Boolean_numeric"



import os

import numpy as np
import pandas as pd
import pytest

from superwise.controller.summary.entities_validator import EntitiesValidationError
from superwise.controller.summary.entities_validator import EntitiesValidator
from superwise.models.task import Task
from superwise.resources.superwise_enums import TaskTypes
from tests.resources.data_entity import DATA_ENTITY_ASSETS_DIRECTORY


@pytest.fixture(scope="function")
def basic_baseline_and_schema():
    basic_baseline = pd.read_json(
        os.path.join(DATA_ENTITY_ASSETS_DIRECTORY, "baseline", "basic_baseline.json"), convert_dates=["label_ts", "ts"]
    )
    basic_schema = pd.read_json(os.path.join(DATA_ENTITY_ASSETS_DIRECTORY, "baseline", "basic_schema.json"))
    yield basic_baseline, basic_schema


def test_schema_has_not_reserved_names(basic_baseline_and_schema):
    baseline, schema = basic_baseline_and_schema
    baseline["task_id"] = "my task id"
    baseline["version_id"] = "my version id"
    schema = schema.append(
        dict(dimension_start_ts=None, name="task_id", role="feature", type="Categorical"), ignore_index=True
    )
    schema = schema.append(
        dict(dimension_start_ts=None, name="version_id", role="feature", type="Categorical"), ignore_index=True
    )
    with pytest.raises(EntitiesValidationError) as validation_error:
        EntitiesValidator(schema, baseline).prepare()
    assert "contains reserved words" in validation_error.value.reason


def test_column_names_match_schema(basic_baseline_and_schema):
    baseline, schema = basic_baseline_and_schema
    schema["name"] = schema["name"].str.upper()
    with pytest.raises(EntitiesValidationError) as validation_error:
        EntitiesValidator(schema, baseline).prepare()
    assert "columns exist in data but are missing" in validation_error.value.reason


def test_has_all_required_roles(basic_baseline_and_schema):
    baseline, schema = basic_baseline_and_schema
    a_baseline, a_schema = baseline.drop("record_id", axis=1), schema[schema["role"] != "id"]
    with pytest.raises(EntitiesValidationError) as validation_error:
        EntitiesValidator(a_schema, a_baseline).prepare()
    assert "columns exist in required roles but are missing from existing roles" in validation_error.value.reason


def test_is_valid_dimension_count(basic_baseline_and_schema):
    baseline, schema = basic_baseline_and_schema
    schema["dimension_start_ts"] = "2020-01-01 00:00:00"
    with pytest.raises(EntitiesValidationError) as validation_error:
        EntitiesValidator(schema, baseline).prepare()
    assert "Maximum allowed dimensions is 10" in validation_error.value.reason


def test_is_valid_dimension(basic_baseline_and_schema):
    baseline, schema = basic_baseline_and_schema
    schema.loc[schema["name"] == "prediction_probability", "dimension_start_ts"] = "2020:01:01 00:00:00"
    with pytest.raises(EntitiesValidationError) as validation_error:
        EntitiesValidator(schema, baseline).prepare()
    assert "type Categorical is required" in validation_error.value.reason


def test_time_stamp_types_valid(basic_baseline_and_schema):
    baseline, schema = basic_baseline_and_schema
    baseline["ts"] = "yesterday"
    with pytest.raises(EntitiesValidationError) as validation_error:
        EntitiesValidator(schema, baseline).prepare()
    assert "ts is marked as type Timestamp but is malformed" in validation_error.value.reason


def test_boolean_types_valid(basic_baseline_and_schema):
    baseline, schema = basic_baseline_and_schema
    baseline["boolean_flag"] = baseline["boolean_flag"].map({True: "Elvis", False: "Presley"})
    with pytest.raises(EntitiesValidationError) as validation_error:
        EntitiesValidator(schema, baseline).prepare()
    assert "Boolean column must be either a subset" in validation_error.value.reason


def test_is_valid_record_id(basic_baseline_and_schema):
    baseline, schema = basic_baseline_and_schema
    baseline["record_id"] = 1
    with pytest.raises(EntitiesValidationError) as validation_error:
        EntitiesValidator(schema, baseline).prepare()
    assert "duplicated values" in validation_error.value.reason


def test_is_valid_prediction_probability(basic_baseline_and_schema):
    baseline, schema = basic_baseline_and_schema
    baseline["prediction_probability"] = baseline["prediction_probability"] * 100
    with pytest.raises(EntitiesValidationError) as validation_error:
        EntitiesValidator(schema, baseline).prepare()
    assert "0 to 1 is expected" in validation_error.value.reason


def test_is_valid_label_ts(basic_baseline_and_schema):
    baseline, schema = basic_baseline_and_schema
    baseline.loc[0, "label_ts"] = pd.NaT
    with pytest.raises(EntitiesValidationError) as validation_error:
        EntitiesValidator(schema, baseline).prepare()
    assert "has null values" in validation_error.value.reason


def test_is_valid_prediction_ts(basic_baseline_and_schema):
    baseline, schema = basic_baseline_and_schema
    baseline.loc[0, "ts"] = pd.NaT
    with pytest.raises(EntitiesValidationError) as validation_error:
        EntitiesValidator(schema, baseline).prepare()
    assert "has null values" in validation_error.value.reason


def test_is_valid_label_weight(basic_baseline_and_schema):
    baseline, schema = basic_baseline_and_schema
    schema = schema.append(
        dict(dimension_start_ts=None, name="label_weight", role="label weight", type="Categorical"), ignore_index=True
    )
    baseline["label_weight"] = "One"
    with pytest.raises(EntitiesValidationError) as validation_error:
        EntitiesValidator(schema, baseline).prepare()
    assert "type Numeric is required" in validation_error.value.reason


def test_happy_flow(basic_baseline_and_schema):
    baseline, schema = basic_baseline_and_schema
    EntitiesValidator(schema, baseline).prepare()


def test_empty_boolean(basic_baseline_and_schema):
    baseline, schema = basic_baseline_and_schema
    baseline["boolean_flag"] = None
    with pytest.raises(EntitiesValidationError) as validation_error:
        EntitiesValidator(schema, baseline).prepare()
    assert "has all null values" in validation_error.value.reason


def test_empty_categorical(basic_baseline_and_schema):
    baseline, schema = basic_baseline_and_schema
    baseline["country"] = None
    with pytest.raises(EntitiesValidationError) as validation_error:
        EntitiesValidator(schema, baseline).prepare()
    assert "has all null values" in validation_error.value.reason


def test_empty_numeric(basic_baseline_and_schema):
    baseline, schema = basic_baseline_and_schema
    baseline["age"] = None
    with pytest.raises(EntitiesValidationError) as validation_error:
        EntitiesValidator(schema, baseline).prepare()
    assert "has all null values" in validation_error.value.reason


def test_empty_ts(basic_baseline_and_schema):
    baseline, schema = basic_baseline_and_schema
    baseline["last_reviewed"] = None
    with pytest.raises(EntitiesValidationError) as validation_error:
        EntitiesValidator(schema, baseline).prepare()
    assert "has all null values" in validation_error.value.reason


def test_regression(basic_baseline_and_schema):

    baseline, schema = basic_baseline_and_schema
    schema.loc[schema["name"].isin(["label", "prediction_value"]), "type"] = "Numeric"
    schema = schema[schema["name"] != "prediction_probability"]
    baseline = baseline.drop("prediction_probability", axis=1)
    baseline["prediction_value"] = np.random.rand(len(baseline))
    baseline["label"] = baseline["prediction_value"]
    EntitiesValidator(schema, baseline).prepare()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("superwise/requirements.txt", "r") as f:
    requirements = [x.strip() for x in f.readlines()]

setuptools.setup(
    name="superwise",
    version="5.0.2",
    description="Superwise SDK",
    url="https://docs.superwise.ai/docs/python-sdk",
    author="Superwise.ai",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author_email="tech@superwise.com",
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=setuptools.find_packages(
        include=[
            "superwise",
            "superwise.models",
            "superwise.controller",
            "superwise.controller.summary",
            "superwise.resources",
        ],
        exclude=[".*"],
    ),
    include_package_data=True,
    zip_safe=False,
    python_requires=">=3",
    setup_requires=[requirements],
    install_requires=[requirements],
)



import logging

import pandas as pd
from scipy.stats import entropy

from .entity_summary_generator import EntitySummaryGenerator
from superwise.resources.superwise_enums import CategoricalSecondaryType


logger = logging.getLogger(__name__)


class CategoricalSummaryGenerator(EntitySummaryGenerator):
    def calc_distribution(self):
        logger.debug(f"Calc distribution for {self._entity.name}")
        buckets = self._entity.value_counts(normalize=True).reset_index()
        buckets.columns = ["value", "frequency"]
        buckets = buckets.sort_values(["frequency", "value"])
        buckets = buckets.to_dict(orient="records")
        entropy = self.calc_entropy(buckets)
        return {"distribution": {"buckets": buckets}, "statistics": {"entropy": entropy}}

    def calc_entropy(self, buckets):
        freq = [b["frequency"] for b in buckets]
        entripy_val = entropy(freq)
        return entripy_val

    def calc_new_values(self):
        logger.debug(f"Calc new values for {self._entity.name}")
        return {"statistics": {"new_values": 0}}

    def calc_unique_values(self):
        logger.debug(f"Calc unique values for {self._entity.name}")
        unique_values = int(len(self._entity.unique()))
        return {"statistics": {"unique_values": unique_values}}

    def calc_top_frequent_percents(self):
        logger.debug(f"Calc top frequent percent for {self._entity.name}")
        top_frequent_percents = float(self._entity.value_counts(normalize=True).sort_values().iloc[-1])
        top_frequent_percents = None if pd.isna(top_frequent_percents) else top_frequent_percents
        return {"statistics": {"top_frequent_percents": top_frequent_percents}}

    def generate_summary(self, secondary_type: str):
        secondary_type_to_metrics = {
            CategoricalSecondaryType.SPARSE.value: [self.calc_missing_values],
            CategoricalSecondaryType.CONSTANT.value: [
                self.calc_missing_values,
                self.calc_new_values,
                self.calc_unique_values,
                self.calc_top_frequent_percents,
                self.calc_distribution,
            ],
            CategoricalSecondaryType.DENSE.value: [
                self.calc_missing_values,
                self.calc_new_values,
                self.calc_unique_values,
                self.calc_top_frequent_percents,
                self.calc_distribution,
            ],
        }

        logger.debug(f"Generate summary for categorical ({secondary_type}) feature {self._entity.name}")
        summary = {}
        for metric in secondary_type_to_metrics[secondary_type]:
            summary = self._update(summary, metric())
        return summary



import json
import os
import sys
from datetime import datetime
from pprint import pprint

import jwt
import pandas as pd
import pytest
import requests
from google.cloud import storage
from google.oauth2 import service_account
from requests import Response

from project_root import PROJECT_ROOT
from superwise import Client
from superwise import Superwise
from superwise.controller.exceptions import *
from superwise.controller.summary.entities_validator import EntitiesValidationError
from superwise.models.data_entity import DataEntity
from superwise.models.task import Task
from superwise.models.version import Version
from superwise.resources.superwise_enums import DataTypesRoles
from superwise.resources.superwise_enums import FeatureType
from superwise.resources.superwise_enums import TaskTypes
from tests import config
from tests import get_entities_fixture
from tests import get_sw


@pytest.fixture(scope="function")
def mock_gcp_client(monkeypatch):
    class GCSClient:
        def __init__(self, *args, **kwargs):
            self.name = "test"

        def bucket(self, bucket_name):
            return GCSClient()

        def blob(self, file_name):
            return GCSClient()

        def download_as_string(self):
            return "asdasdaas"

        def upload_from_string(self, data):
            return None

    monkeypatch.setattr(service_account.Credentials, "from_service_account_info", lambda *args, **kwargs: "")
    monkeypatch.setattr(storage, "Client", lambda *args, **kwargs: GCSClient())


@pytest.fixture(scope="function")
def mock_get_token(monkeypatch):
    monkeypatch.setattr(Client, "get_token", lambda *args, **kwargs: "token")
    monkeypatch.setattr(Client, "get_service_account", lambda *args, **kwargs: {})


@pytest.fixture(scope="function")
def mock_jwt_decode(monkeypatch):
    monkeypatch.setattr(jwt, "decode", lambda *args, **kwargs: {"tenantId": "test"})


@pytest.fixture(scope="function")
def sw(mock_gcp_client, mock_jwt_decode):
    return Superwise(client_id="test", secret="test")


@pytest.fixture(scope="function")
def mock_version_requests(sw, monkeypatch):
    the_response = Response()
    the_response.status_code = 201
    with open(f"{PROJECT_ROOT}/tests/resources/data_entity/assets/entities.json", "r") as fh:
        data_entities = fh.read()
    data_response_response = Response()
    data_response_response._content = data_entities
    data_response_response.status_code = 200
    example_task = Task(name="test", description="test", monitor_delay=1)
    example_version = Version(task_id=1, status="Active", name="test version")

    the_response._content = json.dumps(example_version.get_properties())
    monkeypatch.setattr(requests, "post", lambda *args, **kwargs: the_response)
    monkeypatch.setattr(sw.task, "get_by_id", lambda *args, **kwargs: example_task)
    monkeypatch.setattr(requests, "get", lambda *args, **kwargs: the_response)
    request_get_lambda = lambda url, **kwargs: data_response_response if "data_entities" in url else the_response
    monkeypatch.setattr(requests, "get", request_get_lambda)


def get_entities_fixture(path=f"{PROJECT_ROOT}/tests/resources/basic_schema.json"):
    with open(path, "r") as fh:
        schema = json.loads(fh.read())

    entities = [
        DataEntity(
            dimension_start_ts=m.get("dimension_start_ts", None),
            name=m["name"],
            type=m["type"],
            role=m["role"],
            feature_importance=None,
        )
        for m in schema
    ]
    return entities


def test_dataentity_creation_unit():

    entities = DataEntity(
        dimension_start_ts=None,
        name="barak",
        type=FeatureType.CATEGORICAL,
        role=DataTypesRoles.LABEL,
        feature_importance=None,
    )

    assert entities.type == FeatureType.CATEGORICAL.value
    assert entities.role == DataTypesRoles.LABEL.value


def test_create_version(mock_get_token, sw, mock_version_requests):
    # entities = get_entities_fixture()
    df = pd.read_json(f"{PROJECT_ROOT}/tests/resources/internal_sdk/data_bool.json")
    schema_path = f"{PROJECT_ROOT}/tests/resources/internal_sdk/basic_schema.json"
    with open(schema_path) as f:
        schema = json.load(f)

    roles = {}
    dtypes = {}
    for entity in schema:
        if entity["role"] != "feature":
            roles[entity["name"]] = entity["role"]
        dtypes[entity["name"]] = entity["type"]

    # Example of dtypes
    # dtypes = {
    #     "f0": "Numeric",
    #     "binary_bool_full": "Boolean",
    #     "binary_float_full": "Categorical",
    #     "ts": "Timestamp",
    #     "record_id": "Categorical",
    # }

    # Example of roles
    # roles = {
    #     "f0": "feature",
    #     "binary_bool_full": "feature",
    #     "binary_float_full": "feature",
    #     "ts": "time stamp",
    #     "record_id": "id",
    # }

    entities = sw.data_entity.summarise(
        data=df,
        entities_dtypes=dtypes,
        specific_roles=roles,
        default_role="feature",
        importance=False,
        importance_target_label=None,
        importance_sample=None,
        base_version=None,
    )
    version_external = Version(task_id=1, name="test version", data_entities=entities)

    model = sw.version.create(version_external, wait_until_complete=False)

    assert isinstance(model, Version)
    print(model.get_properties())


def test_create_base_version(mock_get_token, sw, mock_version_requests):

    df = pd.read_json(f"{PROJECT_ROOT}/tests/resources/internal_sdk/data_bool.json")
    schema_path = f"{PROJECT_ROOT}/tests/resources/internal_sdk/basic_schema.json"
    with open(schema_path) as f:
        schema = json.load(f)

    df = df.rename(columns={"f31": "f31_change"})

    roles = {}
    dtypes = {}
    for entity in schema:
        if entity["name"] == "f31":
            entity["name"] = "f31_change"
        if entity["role"] != "feature":
            roles[entity["name"]] = entity["role"]
        dtypes[entity["name"]] = entity["type"]

    base_version = Version(id=1, status="Active")

    entities = sw.data_entity.summarise(
        data=df,
        entities_dtypes=dtypes,
        specific_roles=roles,
        default_role="feature",
        importance=False,
        importance_target_label=None,
        importance_sample=None,
        base_version=base_version,
    )
    version_external = Version(task_id=1, name="test version", data_entities=entities)
    model = sw.version.create(version_external, wait_until_complete=False)
    assert isinstance(model, Version)
    print(model.get_properties())


def test_get_version(mock_get_token, sw, mock_version_requests):
    version = sw.version.get_by_id(1)
    assert isinstance(version, Version)


def test_get_version_data_entities(mock_get_token, sw, mock_version_requests):
    data_entities = sw.version.get_data_entities(version_id=1)
    assert isinstance(data_entities, list)
    assert isinstance(data_entities[0], DataEntity)



import json

import pandas as pd

from .boolean_summary_generator import BooleanSummaryGenerator
from .categorical_summary_generator import CategoricalSummaryGenerator
from .numeric_summary_generator import NumericalSummaryGenerator
from .timestamp_summary_generator import TimestampSummaryGenerator
from .unknown_summary_generator import UnknownSummaryGenerator
from superwise import logger
from superwise.resources.superwise_enums import FeatureType


def get_summary_generator(feature_type):
    return {
        FeatureType.BOOLEAN.value: BooleanSummaryGenerator,
        FeatureType.CATEGORICAL.value: CategoricalSummaryGenerator,
        FeatureType.NUMERIC.value: NumericalSummaryGenerator,
        FeatureType.TIMESTAMP.value: TimestampSummaryGenerator,
        FeatureType.UNKNOWN.value: UnknownSummaryGenerator,
    }[feature_type]


class Summary:
    def __init__(self, entities_df, data):
        self.logger = logger
        self._entities_df = entities_df
        self._data = data.copy()

    def generate(self):
        list_skipped = []

        def summarize_row(row):
            try:
                if not pd.isna(row["id"]):
                    list_skipped.append(str(row["id"]))
                    return ""
                summary_generator = get_summary_generator(row["type"])
                kwargs = dict(entity=self._data[row["name"]])
                return json.loads(
                    pd.Series(summary_generator(**kwargs).generate_summary(row["secondary_type"])).to_json()
                )
            except:
                self.logger.exception(
                    "Failed to generate summary for {}. " f"Setting to null and continuing".format(row["name"]),
                    exc_info=True,
                )
                raise Exception("Exception while summarizing entities row")

        if "summary" not in self._entities_df or self._entities_df["summary"].isnull().any():
            logger.debug("Calculating summary based on data")
            self._entities_df["summary"] = self._entities_df.apply(summarize_row, axis=1)
            if list_skipped:
                logger.info("skipped summary list (already in db): {}".format(",".join(list_skipped)))
            return self._entities_df
        else:
            self.logger.debug("Using user provided entity summary")



""" wrapper for requests class and superwise token handling """
import jwt
import requests

from superwise import logger
from superwise.config import Config
from superwise.controller.exceptions import *


def token_retry(input_func):
    """
    A decorator for token retry handling (ie, refresh token if needed)
    :param input_func:
    :return:
    """

    def wrapper(*args, **kwargs):
        res = input_func(*args, **kwargs)
        if res.status_code == 403:
            args[0].refresh_token()
            res = input_func(*args, **kwargs)
        return res

    return wrapper


class Client:
    """ Client is a wrapper for requests for superwise pacakge """

    def __init__(self, client_id, secret, api_host, email=None, password=None):
        self.client_id = client_id
        self.secret = secret
        self.api_host = api_host
        self.email = email
        self.password = password
        self.token = self.get_token()
        self.tenant_id = self.get_tenant_id(self.token)
        self.logger = logger
        self.headers = self.build_headers()
        self.service_account = self.get_service_account()

    def get_tenant_id(self, token):
        try:
            jwt_payload = jwt.decode(token, "secret", options={"verify_signature": False})
        except:
            raise SuperwiseTokenException("error parsing jwt token: {}".format(token))
        if jwt_payload:
            return jwt_payload.get("tenantId")
        else:
            raise SuperwiseException("error decoding JWT")

    def build_headers(self):
        """
        return the headers to send in each API call
        :return: list of headers
        """
        return {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "Authorization": "Bearer " + self.token,
        }

    def refresh_token(self):
        """
        refresh  bearer token

        :return: token string
        """
        return self.get_token()

    def get_token(self):
        """
        get bearer token to use in each API call

        :return: token string
        """

        headers = {"Accept": "application/json", "Content-Type": "application/json"}
        if self.email:
            url = "{}/identity/resources/auth/v1/user".format(Config.FRONTEGG_URL)
            params = {"email": self.email, "password": self.password}
        else:
            params = {"clientId": self.client_id, "secret": self.secret}
            url = "{}/identity/resources/auth/v1/api-token".format(Config.FRONTEGG_URL)
        res = self._post(url, params=params, headers=headers)

        error = False
        token = None
        try:
            token = res.json().get("accessToken")
        except:
            error = True
        if not token or error:
            raise SuperwiseAuthException("Error get or refresh token url={} params={} ".format(url, params))

        return token

    def _post(self, url, params, headers=None):
        """
        wrapper to requests.post(), with no refresh token handling
        :param url: url string
        :param params: json paramters to send
        :param headers: headers string
        :return: Response object (requests package)
        """
        headers = self.headers if not headers else headers
        logger.debug("POST:  {} params: {}".format(url, params))
        res = requests.post(url, json=params, headers=headers)
        return res

    @token_retry
    def post(self, url, params, headers=None):
        """
        wrapper for self._post() decoarted using token_retry.

        :param url: url string
        :param params: params: json paramters to send
        :param headers: headers: headers string

        :return: Response object (requests package)
        """
        return self._post(url, params, headers)

    @token_retry
    def get(self, url, query_params=None, headers=None):
        """
        wrapper for reuqests.get()

        :param url: url string
        :param query_params:  dictionary of paramters to add as query string
        :param headers: headers dictionary
        :return: Response object (requests package)
        """
        headers = self.headers if not headers else headers
        logger.debug("GET {} query params: {}".format(url, query_params))
        res = requests.get(url=url, headers=headers, params=query_params)
        return res

    @token_retry
    def delete(self, idx):
        """
        wrapper for reuqests.delete()

        :param idx: id int
        :return: Response object (requests package)
        """
        raise NotImplementedError()

    @token_retry
    def patch(self, url, params, headers=None):
        """
        wrapper for reuqests.delete()

        :return: Response object (requests package)
        """
        headers = self.headers if not headers else headers
        logger.debug("PATCH {}  params: {} ".format(url, params))
        res = requests.patch(url, json=params, headers=headers)
        return res

    def get_service_account(self):
        url = "https://{}/{}/{}".format(self.api_host, self.tenant_id, "admin/v1/settings/service-account")
        logger.debug("GET Service Account")
        try:
            res = requests.get(url, headers=self.headers)
            service_account = res.json().get("service_account")
        except Exception as e:
            logger.debug("Failed GET Service Account")
            raise SuperwiseServiceAccountException("Failed get service account with ext {}".format(e))
        return service_account



""" This module implement validation errors model  """
from superwise.controller.exceptions import *
from superwise.models.base import BaseModel


class ValidationError(BaseModel):
    """ validation error model class, model  for validation errors """

    def __init__(self, http_status_code=None, http_error_reason=None, body=None):
        """

        :param http_status_code: http status code from server
        :param http_error_reason: reason for error
        :param body:jbody of response object
        """
        self.http_status_code = http_status_code
        self.http_error_reason = http_error_reason
        self.details = []
        if isinstance(body, str):
            self.error = body
        else:
            detail = body.get("detail")
            if detail:
                self.error = "Input Validation error"
                if isinstance(detail, list):
                    for item in detail:
                        self.details.append(
                            {
                                "field": ".".join(item["loc"]),
                                "error": "{} in field {}".format(item["msg"], item["loc"][1]),
                            }
                        )
                else:
                    self.details = detail
        raise SuperwiseValidationException(self.get_properties())



""" root for supwrwise pacakge, set logger and load config"""
import json
import logging
import os
import pkgutil

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

vcr_log = logging.getLogger("vcr")
vcr_log.setLevel(logging.ERROR)

from superwise.config import Config
from superwise.controller.client import Client
from superwise.controller.task import TaskController
from superwise.controller.version import VersionController
from superwise.controller.role import RoleController
from superwise.controller.dataentity import DataEntityController
from superwise.controller.transaction import TransactionController
from superwise.controller.segment import SegmentController


class Superwise:
    """ Superwise class - main class for superwise package """

    def __init__(
        self,
        client_id=None,
        secret=None,
        _rest_client=None,
        email=None,
        password=None,
        _fegg_url=None,
        _superwise_host=None,
    ):
        """
        constructer for Superwise class

        :param client_id:
        :param secret:
        :param _rest_client: inject rest client if needed (allow mocking of rest api calls)
        """

        self.logger = logger
        if _superwise_host:
            Config.SUPERWISE_HOST = _superwise_host
        if _fegg_url:
            Config.FRONTEGG_URL = _fegg_url
        client_id = client_id or os.environ.get("SUPERWISE_CLIENT_ID")
        secret = secret or os.environ.get("SUPERWISE_SECRET")
        if email and password:
            self.logger.info("login using user and password")
        elif secret is None or client_id is None:
            raise Exception("secret or email/password are mendatory fields")
        api_host = Config.SUPERWISE_HOST
        if not _rest_client:
            _rest_client = Client(client_id, secret, api_host, email, password)
        self.tenant_id = _rest_client.tenant_id
        self.task = TaskController(_rest_client, self)
        self.version = VersionController(_rest_client, self)
        self.role = RoleController(_rest_client, self)
        self.data_entity = DataEntityController(_rest_client, self)
        self.transaction = TransactionController(_rest_client, self)
        self.segment = SegmentController(_rest_client, self)



""" This module implement data functionality  """
import json
import re
from typing import List
from typing import Optional

import boto3
import google
from google.cloud import storage
from google.oauth2 import service_account as GCPServiceAccount

from superwise import Config
from superwise.controller.base import BaseController
from superwise.controller.exceptions import SuperwiseException
from superwise.controller.exceptions import SuperwiseStorageDownloadGCSError
from superwise.controller.exceptions import SuperwiseStorageDownloadS3Error
from superwise.controller.exceptions import SuperwiseStorageUploadGCSError
from superwise.controller.exceptions import SuperwiseValidationException
from superwise.models.task import Task


class TransactionController(BaseController):
    """Transaction Controller is in-charge for create transaction using file and batch request """

    def __init__(self, client, sw):
        """
        constructer for DataController class

        :param client:

        """
        super().__init__(client, sw)
        self.path = "gateway/v1/transaction"
        self.model_name = None
        _bucket_name = "superwise-{}-{}".format(self.client.tenant_id, Config.ENVIRONMENT)
        self._gcs_internal_bucket = self._create_gcs_bucket_connection(
            bucket_name=_bucket_name, service_account=self.client.service_account
        )

    def _create_gcs_bucket_connection(self, bucket_name, service_account):
        try:
            self.logger.debug(f"Create connection to superwise bucket {bucket_name}")
            credentials = GCPServiceAccount.Credentials.from_service_account_info(service_account)
            gcs_client = storage.Client(credentials=credentials)
            return gcs_client.bucket(bucket_name)
        except Exception as e:
            self.logger.error(f"Error create connection to superwise bucket {bucket_name}")
            raise Exception(f"Error create connection to superwise bucket {bucket_name}")

    def _extract_directory(self, url: str):
        bucket = url.split("/")[2]
        prefix = "/".join(url.split("/")[3:])
        return bucket, prefix

    def _upload_string_to_internal_bucket(self, data, file_name):
        try:
            self.logger.debug(f"Upload file to superwise bucket {file_name}")
            blob = self._gcs_internal_bucket.blob(f"landing/{file_name}")
            blob.upload_from_string(data=data)
            return f"gs://{self._gcs_internal_bucket.name}/{file_name}"
        except google.api_core.exceptions.Forbidden as e:
            if "does not have storage.objects.delete access" in e.message:
                self.logger.error(f"Failed upload file to superwise bucket because {file_name} already exist")
                raise SuperwiseStorageUploadGCSError(
                    f"Failed upload file to superwise bucket because {file_name} already exist"
                )
            raise SuperwiseStorageUploadGCSError(f"Failed upload file to superwise storage {file_name}")
        except Exception as e:
            self.logger.error(f"Failed upload file to superwise storage {file_name}")
            raise SuperwiseStorageUploadGCSError(f"Failed upload file to superwise storage {file_name}")

    def log_batch(self, task_id: str, records: List[dict], version_id: Optional[str] = None):
        """
        stream data of a given file path

        :param
        - task_id: string - model which the data associated to him.
        - version_id: string - version of the model -   Optional
        - records: List[dict] - list of records of data,  each record is a dict.
        :return transaction_id
        """
        self.logger.info("transaction batch")
        payload = dict(records=records, task_id=task_id)
        if version_id is not None:
            payload["version_id"] = version_id
        r = self.client.post(self.build_url("{}".format(self.path + "/batch")), payload)
        self.logger.info("file_log server response: {}".format(r.content))
        if r.status_code == 201:
            return r.json()
        else:
            raise Exception("send records to superwise failed, server error")

    def _create_s3_client(self, aws_access_key_id, aws_secret_access_key, aws_session_token, role_arn):
        try:
            if role_arn is not None:
                self.logger.debug("Create S3 client from role_arn")
                sts_client = boto3.client("sts")
                assumed_role_object = sts_client.assume_role(RoleArn=role_arn, RoleSessionName="superwise-session")
                credentials = assumed_role_object["Credentials"]

                return boto3.client(
                    "s3",
                    aws_access_key_id=credentials["AccessKeyId"],
                    aws_secret_access_key=credentials["SecretAccessKey"],
                    aws_session_token=credentials["SessionToken"],
                )
            elif aws_access_key_id is not None and aws_secret_access_key is not None:
                self.logger.debug("Create S3 client with aws access key and secret key")
                return boto3.client(
                    "s3",
                    aws_access_key_id=aws_access_key_id,
                    aws_secret_access_key=aws_secret_access_key,
                    aws_session_token=aws_session_token,
                )

            else:
                self.logger.error("Error create s3 client - role arn or access and secret keys not provided")
                raise
        except Exception as e:
            self.logger.error(f"Error create s3 client, ext {e}")
            raise SuperwiseException("Error create s3 client - role arn or access and secret keys not provided")

    def log_from_s3(
        self, file_path, aws_access_key_id=None, aws_secret_access_key=None, role_arn=None, aws_session_token=None
    ):
        if not str(file_path).startswith("s3://"):
            self.logger.error(f"Failed upload file to superwise storage {file_path}")
            raise Exception("file_path must start with 's3://'")
        try:
            self.logger.info("Download file {} from s3".format(file_path))
            bucket, key = self._extract_directory(file_path)
            s3_client = self._create_s3_client(
                aws_access_key_id=aws_access_key_id,
                aws_secret_access_key=aws_secret_access_key,
                aws_session_token=aws_session_token,
                role_arn=role_arn,
            )
            resp = s3_client.get_object(Bucket=bucket, Key=key)
            content = resp["Body"]
        except Exception as e:
            self.logger.error(f"Error download file from customer s3 bucket {file_path}")
            raise SuperwiseStorageDownloadS3Error(f"Error download file from customer s3 bucket {file_path}")
        superwise_file_path = self._upload_string_to_internal_bucket(data=content.read(), file_name=key)
        return self.log_file(file_path=superwise_file_path, _origin_path=file_path)

    def log_from_gcs(self, file_path, service_account):
        if not str(file_path).startswith("gs://"):
            self.logger.error("Failed upload file to superwise storage")
            raise Exception("file_path must start with 'gs://'")
        self.logger.info("Download file {} from gcs".format(file_path))
        bucket, key = self._extract_directory(file_path)
        try:
            customer_bucket = self._create_gcs_bucket_connection(bucket, service_account)
            blob = customer_bucket.blob(key)
            data = blob.download_as_string()
        except Exception as e:
            self.logger.error(f"Error download file {file_path} from gcs")
            raise SuperwiseStorageDownloadGCSError(f"Error download file {file_path} from gcs")
        superwise_file_path = self._upload_string_to_internal_bucket(data=data, file_name=key)
        return self.log_file(file_path=superwise_file_path, _origin_path=file_path)

    def log_file(self, file_path, _origin_path=None):
        """
        stream data of a given file path
        :param file_path: url for file stored in cloud str
        :param _origin_path: url for file stored in customer bucket - for internal usage
        :return transaction_id
        """
        self.logger.info("transaction file %s ", file_path)
        pattern = "(s3|gs)://.+"
        if not re.match(pattern, file_path):
            raise SuperwiseValidationException(
                "transaction file failed because of wrong file path. file path should be gcs or s3 path."
            )
        params = {"file": file_path}
        if _origin_path is not None:
            params["origin_path"] = _origin_path
        r = self.client.post(url=self.build_url("{}".format(self.path + "/file")), params=params)
        self.logger.info("transaction file server response: {}".format(r.content))
        if r.status_code == 201:
            return r.json()
        else:
            raise Exception("send file to superwise failed, server error")
